import numpy as np
import sharedmem


class LitPixelCounter:
    """Lit-pixel counter: counts pixels with a signal above a threshold."""

    channel = "litpx"
    output_fields = [
        "cellId", "pulseId", "trainId",
        "litPixels", "unmaskedPixels", "totalIntensity"
    ]
    required_data = [
        "data", "mask", "cellId", "pulseId", "trainId"
    ]

    def __init__(self, data, threshold=0.7):
        """Initialize the instance of lit-pixel analysis addon.

        Parameters
        ----------
        data: dict
            Dictionary with image data. It must include the fields:
            `data`, `mask`, `cellId`, `pulseId`, `trainId`
        threshold: float
            The pixel intensity value, which if exceeded means
            that the pixel is illuminated.
        """
        required_data = set(self.required_data) | {"pulseId"}
        for name in required_data:
            if name not in data:
                raise ValueError(f"The field '{name}' is missed in 'data'")

        self.data = data.copy()
        self.max_images = data["pulseId"].shape[0]
        self.num_images = self.max_images

        # specific members
        self.image = data["data"]
        self.mask = data["mask"]

        self.threshold = threshold
        self.num_unmasked_px = sharedmem.full(self.max_images, 0, int)
        self.num_lit_px = sharedmem.full(self.max_images, 0, int)
        self.total_intensity = sharedmem.full(self.max_images, 0, int)

        self.data["litPixels"] = self.num_lit_px
        self.data["unmaskedPixels"] = self.num_unmasked_px
        self.data["totalIntensity"] = self.total_intensity

    def set_num_images(self, num_images):
        """Sets the actual number of images in data.

        Parameters
        ----------
        num_images: int
            The actual number of images in data
        """
        self.num_images = num_images

    def process(self, chunk):
        """Processes the image data.

        Parameters
        ----------
        chunk: slice, sequence or array
            The indices of images in `data` to process
        """
        ix = range(*chunk.indices(self.num_images))
        for i in ix:
            mask = self.mask[i] == 0
            self.total_intensity[i] = np.sum(
                self.image[i], initial=0, where=mask)
            self.num_lit_px[i] = np.sum(
                self.image[i] > self.threshold, initial=0, where=mask)
            self.num_unmasked_px[i] = np.sum(mask)

    def source_name(self, karabo_id, channel):
        """Returns the source name.

        Parameters
        ----------
        karabo_id: str
            The detector Karabo Id, e.g. SPB_DET_AGIPD1M-1
        channel: str
            The detector channel Id, e.g. 15CH0:output

        Returns
        -------
        source_name: str
            The source name for EXDF files
        """
        return f"{karabo_id}/LITPX/{channel}"

    def create_schema(self, source, file_trains=None, count=None):
        """Creates the indices and keys in the source.

        Parameters
        ----------
        source: InstrumentSource
            The source, where to create the keys
        file_trains: sequence or array
            The list of trains in the file
        count: array
            The count of entry per train for this channel. If None,
            then addons create it using `trainId` field in `data`.
        """
        if file_trains is None:
            file_trains = source.file["INDEX/trainId"][:]

        if count is None:
            tid = self.data["trainId"][:self.num_images]
            trains, count = np.unique(tid, return_counts=True)
            count = count[np.in1d(trains, file_trains)]

        if len(file_trains) != len(count):
            raise ValueError(
                "The length of data count does not match the number of trains")
        if np.sum(count) != self.num_images:
            raise ValueError(
                "The sum of data count does not match "
                "the total number of data entries")

        source.create_index(**{self.channel: count})
        for key in self.output_fields:
            source.create_dataset(
                f"{self.channel}/{key}",
                shape=(self.num_images,),
                dtype=self.data[key].dtype
            )

    def write(self, source):
        """Writes data to file.

        Parameters
        ----------
        source: InstrumentSource
            The source, where to write data
        """
        channel = source[self.channel]
        for key in self.output_fields:
            channel[key][:] = self.data[key][:self.num_images]
